#ifndef IST_BITMAP
#define IST_BITMAP

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include "rgba.h"



namespace ist {

typedef unsigned char u_char;


class Bitmap {
private:
	struct Pixel {
		unsigned int width, height;
		std::vector<bRGBA> data;

		Pixel() : width(0), height(0) {}
		bRGBA* operator [] (int idx) { return &data[width*idx]; }
		const bRGBA* operator [] (int idx) const { return &data[width*idx]; }
	};

	Pixel pixel;

public:
	Bitmap() { }
	explicit Bitmap(const char *filename) { load(filename); }

	bRGBA* operator [](int i) { return pixel[i]; }
	const bRGBA* operator [](int i) const { return pixel[i]; }

	bool empty() const { return pixel.data.empty(); }
	void resize(int width, int height) { pixel.width=width; pixel.height=height; pixel.data.resize(width*height); }
	std::vector<bRGBA>::iterator begin() { return pixel.data.begin(); }
	std::vector<bRGBA>::iterator end() { return pixel.data.end(); }

	bool load(const char *filename);
	bool save(const char *filename) const;

	bool loadBMP(const char *filename);
	bool loadTGA(const char *filename);
	bool saveBMP(const char *filename) const;
	bool saveTGA(const char *filename) const;

	unsigned int getWidth() const { return pixel.width; }
	unsigned int getHeight() const { return pixel.height; }
};


} //ist

#endif
