#include "gl_include.h"
#include "gl_light.h"


namespace ist {
namespace gl {


std::map<int, Light*> Light::light_list;


Light::Light()
	: ambient(0.2, 0.2, 0.2, 1.0),
	  diffuse(0.8, 0.8, 0.8, 1.0),
	  specular(1.0, 1.0, 1.0, 1.0),
	  id(0)
{
	static int light_table[8]={GL_LIGHT0, GL_LIGHT1, GL_LIGHT2, GL_LIGHT3, GL_LIGHT4, GL_LIGHT5, GL_LIGHT6, GL_LIGHT7};

	for(int i=0; i<8; ++i) {
		if(light_list[ light_table[i] ]==0) {
			id=light_table[i];
			light_list[id]=this;
			break;
		}
	}
}
	
Light::~Light()
{
	if(id!=0)
		light_list[id]=0;
}

void Light::enable()
{
	glEnable(id);
	glLightfv(id, GL_POSITION, position.v);
	glLightfv(id, GL_AMBIENT,  ambient.v);
	glLightfv(id, GL_DIFFUSE,  diffuse.v);
	glLightfv(id, GL_SPECULAR, specular.v);
}

void Light::disable()
{
	glDisable(id);
}


} //gl
} //ist

