#include "gl_include.h"
#include "gl_polygon_model.h"

#if defined(ENABLE_GL_EXT) && defined(ENABLE_ARB_VBO)
	#define _ENABLE_ARB_VBO
	#define BUFFER_OFFSET(i) ((char*)0 + (i))
#endif





namespace ist {
namespace gl {

// gl extention

#ifdef _ENABLE_ARB_VBO
PFNGLBINDBUFFERARBPROC glBindBufferARB;
PFNGLBUFFERDATAARBPROC glBufferDataARB;
PFNGLBUFFERSUBDATAARBPROC glBufferSubDataARB;
PFNGLDELETEBUFFERSARBPROC glDeleteBuffersARB;
PFNGLGENBUFFERSARBPROC glGenBuffersARB;
PFNGLGETBUFFERPARAMETERIVARBPROC glGetBufferParameterivARB;
PFNGLGETBUFFERPOINTERVARBPROC glGetBufferPointervARB;
PFNGLGETBUFFERSUBDATAARBPROC glGetBufferSubDataARB;
PFNGLISBUFFERARBPROC glIsBufferARB;
PFNGLUNMAPBUFFERARBPROC glUnmapBufferARB;
#endif


bool ARB_vertex_buffer_object::is_supported=false;

bool ARB_vertex_buffer_object::enable_=true;

ARB_vertex_buffer_object::ARB_vertex_buffer_object()
{
#ifdef _ENABLE_ARB_VBO
	glBindBufferARB = (PFNGLBINDBUFFERARBPROC)glGetProcAddress("glBindBufferARB");
	glBufferDataARB = (PFNGLBUFFERDATAARBPROC)glGetProcAddress("glBufferDataARB");
	glBufferSubDataARB = (PFNGLBUFFERSUBDATAARBPROC)glGetProcAddress("glBufferSubDataARB");
	glDeleteBuffersARB = (PFNGLDELETEBUFFERSARBPROC)glGetProcAddress("glDeleteBuffersARB");
	glGenBuffersARB = (PFNGLGENBUFFERSARBPROC)glGetProcAddress("glGenBuffersARB");
	glGetBufferParameterivARB = (PFNGLGETBUFFERPARAMETERIVARBPROC)glGetProcAddress("glGetBufferParameterivARB");
	glGetBufferPointervARB = (PFNGLGETBUFFERPOINTERVARBPROC)glGetProcAddress("glGetBufferPointervARB");
	glGetBufferSubDataARB = (PFNGLGETBUFFERSUBDATAARBPROC)glGetProcAddress("glGetBufferSubDataARB");
	glIsBufferARB = (PFNGLISBUFFERARBPROC)glGetProcAddress("glIsBufferARB");
	glUnmapBufferARB = (PFNGLUNMAPBUFFERARBPROC)glGetProcAddress("glUnmapBufferARB");

	is_supported=(glBindBufferARB!=0);
#endif
}

ARB_vertex_buffer_object::~ARB_vertex_buffer_object() {}

bool ARB_vertex_buffer_object::isSupported() {
	static ARB_vertex_buffer_object init;
	return ARB_vertex_buffer_object::is_supported;
}

bool ARB_vertex_buffer_object::isEnabled()
{
	return ARB_vertex_buffer_object::enable_;
}



PolygonChunc::PolygonChunc()
	: vbo_id(0), vertex_pointer(0)
{}

PolygonChunc::~PolygonChunc()
{
	delete[] vertex_pointer;

#ifdef _ENABLE_ARB_VBO
	if(vbo_id!=0)
		glDeleteBuffersARB(1, &vbo_id);
#endif
}

void PolygonChunc::setup()
{
	BasePolygonChunc::setup();

	vertex_pointer=new float[ idx_array.size()*8 ];

	int offset=0;
	for(unsigned int i=0; i<idx_array.size(); ++i) {
		vertex_pointer[i*3+0]=vtx_array[ idx_array[i] ].x;
		vertex_pointer[i*3+1]=vtx_array[ idx_array[i] ].y;
		vertex_pointer[i*3+2]=vtx_array[ idx_array[i] ].z;
	}
	offset=idx_array.size()*3;

	for(unsigned int i=0; i<idx_array.size(); ++i) {
		vertex_pointer[offset+i*3+0]=vnml_array[i].x;
		vertex_pointer[offset+i*3+1]=vnml_array[i].y;
		vertex_pointer[offset+i*3+2]=vnml_array[i].z;
	}
	offset=idx_array.size()*6;

	if(!tex_array.empty()) {
		for(unsigned int i=0; i<idx_array.size(); ++i) {
			vertex_pointer[offset+i*2+0]=tex_array[i].x;
			vertex_pointer[offset+i*2+1]=tex_array[i].y;
		}
	}
	else {
		for(unsigned int i=0; i<idx_array.size(); ++i) {
			vertex_pointer[offset+i*2+0]=0;
			vertex_pointer[offset+i*2+1]=0;
		}
	}

#ifdef _ENABLE_ARB_VBO
	if(ARB_vertex_buffer_object::isSupported() && ARB_vertex_buffer_object::isEnabled()) {
		glGenBuffersARB(1, &vbo_id);
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_id);
		glBufferDataARB(GL_ARRAY_BUFFER_ARB, idx_array.size()*8*sizeof(float), vertex_pointer, GL_STATIC_DRAW_ARB);

		delete[] vertex_pointer;
		vertex_pointer=0;
	}
#endif
}


void PolygonChunc::draw() const
{
#ifdef _ENABLE_ARB_VBO
	if(vbo_id!=0) {
		glBindBufferARB(GL_ARRAY_BUFFER_ARB, vbo_id);
		glVertexPointer(3, GL_FLOAT, 0, 0);
		glNormalPointer(GL_FLOAT, 0, BUFFER_OFFSET(idx_array.size()*3*sizeof(float)) );
		glTexCoordPointer(2, GL_FLOAT, 0, BUFFER_OFFSET(idx_array.size()*6*sizeof(float)) );
	}
	else {
		glVertexPointer(3, GL_FLOAT, 0, vertex_pointer);
		glNormalPointer(GL_FLOAT, 0, vertex_pointer+idx_array.size()*3 );
		glTexCoordPointer(2, GL_FLOAT, 0, vertex_pointer+idx_array.size()*6 );
	}
#else
	glVertexPointer(3, GL_FLOAT, 0, vertex_pointer);
	glNormalPointer(GL_FLOAT, 0, vertex_pointer+idx_array.size()*3 );
	glTexCoordPointer(2, GL_FLOAT, 0, vertex_pointer+idx_array.size()*6 );
#endif

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glDrawArrays(GL_TRIANGLES, 0, idx_array.size());
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
}


}	// gl
}	// ist


