#include <sstream>
#include <list>
#include "ist_lua.h"
using namespace std;


namespace ist {
namespace lua {


class Stack
{
public:
	Stack() : _size(0), _capacity(0), cur_buf(0) { }
	Stack(unsigned int __size) : _size(0), _capacity(0), cur_buf(0) { reserve(__size); }

	~Stack() {
		for(std::list<char*>::iterator p=buf_list.begin(); p!=buf_list.end(); ++p)
			free(*p);
	}

	void* push(unsigned int __size) {
		resize(_size+__size);
		return static_cast<void*>(cur_buf+_size-__size);
	}

	void resize(unsigned int __size) {
		_size=__size;

		if(_capacity<_size) {
			if(_capacity==0) _capacity=1;
			while(_capacity<_size) _capacity*=2;
			new_buffer();
		}
	}

	void reserve(unsigned int __size)
	{
		if(_capacity<__size) {
			_capacity=__size;
			new_buffer();
		}
	}

	unsigned int size() { return _size; }
	unsigned int capacity() { return _capacity; }


private:
	Stack(const Stack&);
	Stack& operator =(const Stack&);

	void new_buffer()
	{
		_size=0;
		cur_buf=static_cast<char*>(::malloc(_capacity));
		buf_list.push_back(cur_buf);
	}
	
	unsigned int _size, _capacity;
	char *cur_buf;
	std::list<char*> buf_list;
};


namespace math3d {
	Stack lua_stack(65536);

	void stack_reserve(unsigned int n) { lua_stack.reserve(n); }
	void stack_resize(unsigned int n) { lua_stack.resize(n); }
	unsigned int stack_size() { return lua_stack.size(); }
	unsigned int stack_capacity() { return lua_stack.capacity(); }

	int stack_reserve(lua_State *L) { lua_stack.reserve( static_cast<unsigned int>(lua_tonumber(L, -1)) ); return 0; }
	int stack_resize(lua_State *L)  { lua_stack.resize( static_cast<unsigned int>(lua_tonumber(L, -1)) ); return 0; }
	int stack_size(lua_State *L)    { lua_pushnumber( L, static_cast<lua_Number>(lua_stack.size()) ); return 1; }
	int stack_capacity(lua_State *L){ lua_pushnumber( L, static_cast<lua_Number>(lua_stack.capacity()) ); return 1; }
}





void throw_error(lua_State *L, const char *message)
{
	char buf[1024];
	lua_Debug ld_current, ld_root;

	lua_getstack(L, 0, &ld_current);
	lua_getinfo(L, "n", &ld_current);

	lua_getstack(L, 1, &ld_root);
	lua_getinfo(L, "Sl", &ld_root);

	if(message!=0)
		sprintf(buf, "lua::math3d error at %s file:%s line:%d  %s\n",
			ld_current.name, ld_root.source, ld_root.currentline, message);
	else
		sprintf(buf, "lua::math3d error at %s file:%s line:%d\n",
			ld_current.name, ld_root.source, ld_root.currentline);

	throw Error(buf);
}





template <typename T>
inline T* PushStack() {
	return reinterpret_cast<T*>( math3d::lua_stack.push(sizeof(T)) );
}

template <typename T>
inline int Assign(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_islightuserdata(L, 2))
		throw_error(L, "2Ƃ[U[f[^łKv");
#endif

	*static_cast<T*>(lua_touserdata(L, 1)) = *static_cast<T*>(lua_touserdata(L, 2));
	return 0;
}


template <typename T>
inline int GetElem(lua_State *L, int I) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=1)
		throw_error(L, "1ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1))
		throw_error(L, "̓[Uf[^łKv");
#endif

	lua_pushnumber( L, (*static_cast<T*>(lua_touserdata(L, 1)))[I] );
	return 1;
}

template <typename T>
inline int SetElem(lua_State *L, int I) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_isnumber(L, 2))
		throw_error(L, "1̓[U[f[^A2͎łKv");
#endif

	(*static_cast<T*>(lua_touserdata(L, 1)))[I] = lua_tonumber(L, 2);
	return 0;
}

template <typename T>
inline int VecOpAS(lua_State *L, T (T::*op)(const T&) const) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_islightuserdata(L, 2))
		throw_error(L, "2Ƃ[U[f[^łKv");
#endif

	T *p=PushStack<T>();
	*p= (static_cast<T*>(lua_touserdata(L, 1))->*op)( *static_cast<T*>(lua_touserdata(L, 2)) );
	lua_pushlightuserdata( L, static_cast<void*>(p) );
	return 1;
}

template <typename T>
inline int VecOpMD(lua_State *L, T (T::*op)(const T&) const, T (T::*opf)(float) const) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !(lua_islightuserdata(L, 2) || lua_isnumber(L, 2)))
		throw_error(L, "1̓[U[f[^A2͎[U[f[^łKv");
#endif

	T *p=PushStack<T>();
	if( lua_isnumber(L, 2) )
		*p= (static_cast<T*>(lua_touserdata(L, 1))->*opf)(lua_tonumber(L, 2));
	else if( lua_islightuserdata(L, 2) )
		*p= (static_cast<T*>(lua_touserdata(L, 1))->*op)(*static_cast<T*>(lua_touserdata(L, 2)));

	lua_pushlightuserdata( L, static_cast<void*>(p) );
	return 1;
}

template <typename T>
inline int VecOpNorm(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=1)
		throw_error(L, "1ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1))
		throw_error(L, "̓[U[f[^łKv");
#endif

	lua_pushnumber( L, static_cast<T*>(lua_touserdata(L, 1))->norm() );
	return 1;
}

template <typename T>
inline int VecOpNormalize(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=1)
		throw_error(L, "1ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1))
		throw_error(L, "̓[U[f[^łKv");
#endif

	static_cast<T*>(lua_touserdata(L, 1))->normalize();
	lua_pushlightuserdata( L, lua_touserdata(L, 1) );
	return 1;
}

template <typename T>
inline int VecOpDot(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_islightuserdata(L, 2))
		throw_error(L, "2Ƃ[U[f[^łKv");
#endif

	lua_pushnumber( L, static_cast<T*>(lua_touserdata(L, 1))->dot(*static_cast<T*>(lua_touserdata(L, 2))) );
	return 1;
}

template <typename T>
inline int VecOpCross(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_islightuserdata(L, 2))
		throw_error(L, "2Ƃ[U[f[^łKv");
#endif

	T *p=PushStack<T>();
	*p= T().cross( *static_cast<T*>(lua_touserdata(L, 1)), *static_cast<T*>(lua_touserdata(L, 2)) );
	lua_pushlightuserdata( L, static_cast<void*>(p) );
	return 1;
}

namespace lua_vector2 {
	int _new(lua_State *L)
	{
		vector2 *p=PushStack<vector2>();

		if(lua_gettop(L)==1)
			*p=vector2(*static_cast<vector2*>(lua_touserdata(L, 1)));
		else if(lua_gettop(L)==2)
			*p=vector2(lua_tonumber(L, 1), lua_tonumber(L, 2));
		else
			*p=vector2();

		lua_pushlightuserdata( L, static_cast<void*>(p) );
		return 1;
	}

	int split(lua_State *L) {
		vector2 *vec=static_cast<vector2*>(lua_touserdata(L,1));
		lua_pushnumber(L, vec->v[0]);
		lua_pushnumber(L, vec->v[1]);
		return 2;
	}

	int assign(lua_State *L){ return Assign<vector2>(L); }
	int set_x(lua_State *L) { return SetElem<vector2>(L, 0); }
	int set_y(lua_State *L) { return SetElem<vector2>(L, 1); }
	int get_x(lua_State *L) { return GetElem<vector2>(L, 0); }
	int get_y(lua_State *L) { return GetElem<vector2>(L, 1); }
	
	int add(lua_State *L) { return VecOpAS<vector2>(L, &vector2::operator+); }
	int sub(lua_State *L) { return VecOpAS<vector2>(L, &vector2::operator-); }
	int mul(lua_State *L) { return VecOpMD<vector2>(L, &vector2::operator*, &vector2::operator*); }
	int div(lua_State *L) { return VecOpMD<vector2>(L, &vector2::operator/, &vector2::operator/); }
	
	int norm(lua_State *L)      { return VecOpNorm<vector2>(L); }
	int normalize(lua_State *L) { return VecOpNormalize<vector2>(L); }
	int dot(lua_State *L)       { return VecOpDot<vector2>(L); }
}

namespace lua_vector3 {
	int _new(lua_State *L)
	{
		vector3 *p=PushStack<vector3>();

		if(lua_gettop(L)==1)
			*p=vector3(*static_cast<vector3*>(lua_touserdata(L, 1)));
		else if(lua_gettop(L)==3)
			*p=vector3(lua_tonumber(L, 1), lua_tonumber(L, 2), lua_tonumber(L, 3));
		else
			*p=vector3();

		lua_pushlightuserdata( L, static_cast<void*>(p) );
		return 1;
	}

	int split(lua_State *L) {
		vector3 *vec=static_cast<vector3*>(lua_touserdata(L, 1));
		lua_pushnumber(L, vec->v[0]);
		lua_pushnumber(L, vec->v[1]);
		lua_pushnumber(L, vec->v[2]);
		return 3;
	}
	
	int assign(lua_State *L){ return Assign<vector3>(L); }
	int set_x(lua_State *L) { return SetElem<vector3>(L, 0); }
	int set_y(lua_State *L) { return SetElem<vector3>(L, 1); }
	int set_z(lua_State *L) { return SetElem<vector3>(L, 2); }
	int get_x(lua_State *L) { return GetElem<vector3>(L, 0); }
	int get_y(lua_State *L) { return GetElem<vector3>(L, 1); }
	int get_z(lua_State *L) { return GetElem<vector3>(L, 2); }

	int add(lua_State *L) { return VecOpAS<vector3>(L, &vector3::operator+); }
	int sub(lua_State *L) { return VecOpAS<vector3>(L, &vector3::operator-); }
	int mul(lua_State *L) { return VecOpMD<vector3>(L, &vector3::operator*, &vector3::operator*); }
	int div(lua_State *L) { return VecOpMD<vector3>(L, &vector3::operator/, &vector3::operator/); }

	int norm(lua_State *L)      { return VecOpNorm<vector3>(L); }
	int normalize(lua_State *L) { return VecOpNormalize<vector3>(L); }
	int dot(lua_State *L)       { return VecOpDot<vector3>(L); }
	int cross(lua_State *L)     { return VecOpCross<vector3>(L); }
}

namespace lua_vector4 {
	int _new(lua_State *L)
	{
		vector4 *p=PushStack<vector4>();

		if(lua_gettop(L)==1)
			*p=vector4(*static_cast<vector4*>(lua_touserdata(L, 1)));
		else if(lua_gettop(L)>=3) {
			if(lua_gettop(L)==3)
				*p=vector4(lua_tonumber(L, 1), lua_tonumber(L, 2), lua_tonumber(L, 3));
			else if(lua_gettop(L)==4)
				*p=vector4(lua_tonumber(L, 1), lua_tonumber(L, 2), lua_tonumber(L, 3), lua_tonumber(L, 4) );
		}
		else
			*p=vector4();

		lua_pushlightuserdata( L, static_cast<void*>(p) );
		return 1;
	}

	int split(lua_State *L) {
		vector4 *vec=static_cast<vector4*>(lua_touserdata(L, 1));
		lua_pushnumber(L, vec->v[0]);
		lua_pushnumber(L, vec->v[1]);
		lua_pushnumber(L, vec->v[2]);
		lua_pushnumber(L, vec->v[3]);
		return 4;
	}
	
	int assign(lua_State *L){ return Assign<vector4>(L); }
	int set_x(lua_State *L) { return SetElem<vector4>(L, 0); }
	int set_y(lua_State *L) { return SetElem<vector4>(L, 1); }
	int set_z(lua_State *L) { return SetElem<vector4>(L, 2); }
	int set_w(lua_State *L) { return SetElem<vector4>(L, 3); }
	int get_x(lua_State *L) { return GetElem<vector4>(L, 0); }
	int get_y(lua_State *L) { return GetElem<vector4>(L, 1); }
	int get_z(lua_State *L) { return GetElem<vector4>(L, 2); }
	int get_w(lua_State *L) { return GetElem<vector4>(L, 3); }
	
	int add(lua_State *L) { return VecOpAS<vector4>(L, &vector4::operator+); }
	int sub(lua_State *L) { return VecOpAS<vector4>(L, &vector4::operator-); }
	int mul(lua_State *L) { return VecOpMD<vector4>(L, &vector4::operator*, &vector4::operator*); }
	int div(lua_State *L) { return VecOpMD<vector4>(L, &vector4::operator/, &vector4::operator/); }

	int norm(lua_State *L)      { return VecOpNorm<vector4>(L); }
	int normalize(lua_State *L) { return VecOpNormalize<vector4>(L); }
	int dot(lua_State *L)       { return VecOpDot<vector4>(L); }
	int cross(lua_State *L)     { return VecOpCross<vector4>(L); }
}







template <typename T>
inline int MatOpNew(lua_State *L) {
	T *p=PushStack<T>();
	if(lua_gettop(L)==1) { *p=T(*static_cast<T*>(lua_touserdata(L, 1))); }
	else { *p=T(); }
	lua_pushlightuserdata( L, static_cast<void*>(p) );
	return 1;
}

template <typename T>
inline int MatOpMul(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_islightuserdata(L, 2))
		throw_error(L, "2Ƃ[U[f[^łKv");
#endif

	T *p=PushStack<T>();
	*p=(*static_cast<T*>(lua_touserdata(L, 1))) * (*static_cast<T*>(lua_touserdata(L, 2)));
	lua_pushlightuserdata( L, static_cast<void*>(p) );
	return 1;
}

template <class matT, class vecT>
inline int MatOpMulVec(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_islightuserdata(L, 2))
		throw_error(L, "2Ƃ[U[f[^łKv");
#endif

	vecT *p=PushStack<vecT>();
	*p=(*static_cast<matT*>(lua_touserdata(L, 1))) * (*static_cast<vecT*>(lua_touserdata(L, 2)));
	lua_pushlightuserdata( L, static_cast<void*>(p) );
	return 1;
}

template <typename T>
inline int MatOpIdentity(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=1)
		throw_error(L, "1ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1))
		throw_error(L, "̓[U[f[^łKv");
#endif

	static_cast<T*>(lua_touserdata(L, 1))->identity();
	lua_pushlightuserdata( L, lua_touserdata(L, 1) );
	return 1;
}

template <typename T>
inline int MatOpInvert(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=1)
		throw_error(L, "1ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1))
		throw_error(L, "̓[U[f[^łKv");
#endif

	static_cast<T*>(lua_touserdata(L, 1))->invert();
	lua_pushlightuserdata( L, lua_touserdata(L, 1) );
	return 1;
}

template <typename T>
inline int Affine3(lua_State *L, T& (T::*trans)(float,float,float), T& (T::*transv)(const float*) ) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)<=1)
		throw_error(L, "sĂB");
	else if(!lua_islightuserdata(L, 1))
		throw_error(L, "1ڂ̈̓[U[f[^łKv");
#endif

	if(lua_isuserdata(L, 2)) {
		(static_cast<T*>(lua_touserdata(L, 1))->*transv)( static_cast<vector3*>(lua_touserdata(L, 2))->v );
		lua_pushlightuserdata( L, lua_touserdata(L, 1) );
	}
	else if(lua_gettop(L)==4){
		(static_cast<T*>(lua_touserdata(L, 1))->*trans)( lua_tonumber(L, 2), lua_tonumber(L, 3), lua_tonumber(L, 4) );
		lua_pushlightuserdata( L, lua_touserdata(L, 1) );
	}
	else {
		lua_pushnil( L );
#ifndef NO_EXEPTION
		throw_error(L, "2߈ȍ~̈́A1̃xNgA3̎łKv");
#endif
	}
	return 1;
}

template <typename T>
inline int RotateXYZ(lua_State *L, T& (T::*rotf)(float) ) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_isnumber(L, 2))
		throw_error(L, "1̓[U[f[^A2͎łKv");
#endif

	(static_cast<T*>(lua_touserdata(L, 1))->*rotf)( lua_tonumber(L, 2) );
	lua_pushlightuserdata( L, lua_touserdata(L, 1) );
	return 1;
}

template <typename T>
inline int RotateA(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=3)
		throw_error(L, "3ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_islightuserdata(L, 2) || !lua_isnumber(L, 3))
		throw_error(L, "1Ƒ2̓[U[f[^A3͎łKv");
#endif

	static_cast<T*>(lua_touserdata(L,1))->rotateA( *static_cast<vector3*>(lua_touserdata(L, 2)), lua_tonumber(L, 3) );
	lua_pushlightuserdata( L, lua_touserdata(L, 1) );
	return 1;
}

template <typename T>
inline int AimVector(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_islightuserdata(L, 2))
		throw_error(L, "2Ƃ[U[f[^łKv");
#endif

	static_cast<T*>(lua_touserdata(L,1))->aimVector( *static_cast<vector3*>(lua_touserdata(L, 2)) );
	lua_pushlightuserdata( L, lua_touserdata(L, 1) );
	return 1;
}

inline int Mat22OpScale(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=3)
		throw_error(L, "3ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1))
		throw_error(L, "1̓[U[f[^AȊO͎łKv");
#endif

	static_cast<matrix22*>(lua_touserdata(L, 1))->scale( lua_tonumber(L, 2), lua_tonumber(L, 3) );\
	lua_pushlightuserdata( L, lua_touserdata(L, 1) );\
	return 1;
}


inline int Mat22OpRotate(lua_State *L) {
#ifndef NO_EXEPTION
	if(lua_gettop(L)!=2)
		throw_error(L, "2ł͂Ȃ");
	else if(!lua_islightuserdata(L, 1) || !lua_isnumber(L, 2))
		throw_error(L, "1̓[U[f[^A2͎łKv");
#endif

	static_cast<matrix22*>(lua_touserdata(L, 1))->rotate( lua_tonumber(L, 2) );\
	lua_pushlightuserdata( L, lua_touserdata(L, 1) );\
	return 1;
}


namespace lua_matrix22 {
	int _new(lua_State *L)    { return MatOpNew<matrix22>(L); }
	int assign(lua_State *L)  { return Assign<matrix22>(L); }
	int mul(lua_State *L)     { return MatOpMul<matrix22>(L); }
	int mulvec(lua_State *L)  { return MatOpMulVec<matrix22, vector2>(L); }
	int identity(lua_State *L){ return MatOpIdentity<matrix22>(L); }
	int invert(lua_State *L)  { return MatOpInvert<matrix22>(L); }
	int scale(lua_State *L)   { return Mat22OpScale(L); }
	int rotate(lua_State *L)  { return Mat22OpRotate(L); }
}

namespace lua_matrix33 {
	int _new(lua_State *L)    { return MatOpNew<matrix33>(L); }
	int assign(lua_State *L)  { return Assign<matrix33>(L); }
	int mul(lua_State *L)     { return MatOpMul<matrix33>(L); }
	int mulvec(lua_State *L)  { return MatOpMulVec<matrix33, vector3>(L); }
	int identity(lua_State *L){ return MatOpIdentity<matrix33>(L); }
	int invert(lua_State *L)  { return MatOpInvert<matrix33>(L); }

	int scale(lua_State *L)   { return Affine3<matrix33>(L, &matrix33::scale, &matrix33::scale); }
	int rotate(lua_State *L)  { return Affine3<matrix33>(L, &matrix33::rotate, &matrix33::rotate); }
	int rotateX(lua_State *L) { return RotateXYZ<matrix33>(L, &matrix33::rotateX); }
	int rotateY(lua_State *L) { return RotateXYZ<matrix33>(L, &matrix33::rotateY); }
	int rotateZ(lua_State *L) { return RotateXYZ<matrix33>(L, &matrix33::rotateZ); }
	int rotateA(lua_State *L) { return RotateA<matrix33>(L); }
	int aim_vector(lua_State *L) { return AimVector<matrix33>(L); }
}

namespace lua_matrix44 {
	int _new(lua_State *L)     { return MatOpNew<matrix44>(L); }
	int assign(lua_State *L)   { return Assign<matrix44>(L); }
	int mul(lua_State *L)      { return MatOpMul<matrix44>(L); }
	int mulvec(lua_State *L)   { return MatOpMulVec<matrix44, vector4>(L); }
	int identity(lua_State *L) { return MatOpIdentity<matrix44>(L); }
	int invert(lua_State *L)   { return MatOpInvert<matrix44>(L); }

	int translate(lua_State *L){ return Affine3<matrix44>(L, &matrix44::translate, &matrix44::translate); }
	int scale(lua_State *L)    { return Affine3<matrix44>(L, &matrix44::scale, &matrix44::scale); }
	int rotate(lua_State *L)   { return Affine3<matrix44>(L, &matrix44::rotate, &matrix44::rotate); }
	int rotateX(lua_State *L)  { return RotateXYZ<matrix44>(L, &matrix44::rotateX); }
	int rotateY(lua_State *L)  { return RotateXYZ<matrix44>(L, &matrix44::rotateY); }
	int rotateZ(lua_State *L)  { return RotateXYZ<matrix44>(L, &matrix44::rotateZ); }
	int rotateA(lua_State *L)  { return RotateA<matrix44>(L); }
	int aim_vector(lua_State *L) { return AimVector<matrix44>(L); }
}




template <typename T>
inline void PushData(lua_State *L, const T &d) {
	T *p=PushStack<T>();
	*p=d;
	lua_pushlightuserdata( L, static_cast<void*>(p) );

}

template <typename T>
inline T& ToData(lua_State *L, int index) {
	return *static_cast<T*>(lua_touserdata(L, index));

}

void pushvector2(lua_State *L, const vector2 &d) { PushData<vector2>(L, d); }

void pushvector3(lua_State *L, const vector3 &d) { PushData<vector3>(L, d); }
void pushvector4(lua_State *L, const vector4 &d) { PushData<vector4>(L, d); }
void pushmatrix22(lua_State *L, const matrix22 &d) { PushData<matrix22>(L, d); }
void pushmatrix33(lua_State *L, const matrix33 &d) { PushData<matrix33>(L, d); }
void pushmatrix44(lua_State *L, const matrix44 &d) { PushData<matrix44>(L, d); }

vector2& tovector2(lua_State *L, int index) { return ToData<vector2>(L, index); }

vector3& tovector3(lua_State *L, int index) { return ToData<vector3>(L, index); }

vector4& tovector4(lua_State *L, int index) { return ToData<vector4>(L, index); }

matrix22& tomatrix22(lua_State *L, int index) { return ToData<matrix22>(L, index); }
matrix33& tomatrix33(lua_State *L, int index) { return ToData<matrix33>(L, index); }
matrix44& tomatrix44(lua_State *L, int index) { return ToData<matrix44>(L, index); }


inline void RegistFunc(lua_State *L, const char *func_name, int (*func)(lua_State*)) {
	lua_pushstring(L, func_name);
	lua_pushcfunction(L, func);
	lua_settable(L, -3);
}


void math3dlibopen(lua_State *L)
{
	lua_pushstring(L, "math3d");
	lua_newtable(L);
	RegistFunc(L, "reserve", math3d::stack_reserve);
	RegistFunc(L, "resize", math3d::stack_resize);
	RegistFunc(L, "size", math3d::stack_size);
	RegistFunc(L, "capacity", math3d::stack_capacity);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "vector2");
	lua_newtable(L);
	RegistFunc(L, "new", lua_vector2::_new);
	RegistFunc(L, "assign", lua_vector2::assign);
	RegistFunc(L, "split", lua_vector2::split);
	RegistFunc(L, "set_x", lua_vector2::set_x);
	RegistFunc(L, "set_y", lua_vector2::set_y);
	RegistFunc(L, "get_x", lua_vector2::get_x);
	RegistFunc(L, "get_y", lua_vector2::get_y);
	RegistFunc(L, "add", lua_vector2::add);
	RegistFunc(L, "sub", lua_vector2::sub);
	RegistFunc(L, "mul", lua_vector2::mul);
	RegistFunc(L, "div", lua_vector2::div);
	RegistFunc(L, "norm", lua_vector2::norm);
	RegistFunc(L, "normalize", lua_vector2::normalize);
	RegistFunc(L, "dot", lua_vector2::dot);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "vector3");
	lua_newtable(L);
	RegistFunc(L, "new", lua_vector3::_new);
	RegistFunc(L, "assign", lua_vector3::assign);
	RegistFunc(L, "split", lua_vector3::split);
	RegistFunc(L, "set_x", lua_vector3::set_x);
	RegistFunc(L, "set_y", lua_vector3::set_y);
	RegistFunc(L, "set_z", lua_vector3::set_z);
	RegistFunc(L, "get_x", lua_vector3::get_x);
	RegistFunc(L, "get_y", lua_vector3::get_y);
	RegistFunc(L, "get_z", lua_vector3::get_z);
	RegistFunc(L, "add", lua_vector3::add);
	RegistFunc(L, "sub", lua_vector3::sub);
	RegistFunc(L, "mul", lua_vector3::mul);
	RegistFunc(L, "div", lua_vector3::div);
	RegistFunc(L, "norm", lua_vector3::norm);
	RegistFunc(L, "normalize", lua_vector3::normalize);
	RegistFunc(L, "dot", lua_vector3::dot);
	RegistFunc(L, "cross", lua_vector3::cross);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "vector4");
	lua_newtable(L);
	RegistFunc(L, "new", lua_vector4::_new);
	RegistFunc(L, "assign", lua_vector4::assign);
	RegistFunc(L, "split", lua_vector4::split);
	RegistFunc(L, "set_x", lua_vector4::set_x);
	RegistFunc(L, "set_y", lua_vector4::set_y);
	RegistFunc(L, "set_z", lua_vector4::set_z);
	RegistFunc(L, "set_w", lua_vector4::set_w);
	RegistFunc(L, "get_x", lua_vector4::get_x);
	RegistFunc(L, "get_y", lua_vector4::get_y);
	RegistFunc(L, "get_z", lua_vector4::get_z);
	RegistFunc(L, "get_w", lua_vector4::get_w);
	RegistFunc(L, "add", lua_vector4::add);
	RegistFunc(L, "sub", lua_vector4::sub);
	RegistFunc(L, "mul", lua_vector4::mul);
	RegistFunc(L, "div", lua_vector4::div);
	RegistFunc(L, "norm", lua_vector4::norm);
	RegistFunc(L, "normalize", lua_vector4::normalize);
	RegistFunc(L, "dot", lua_vector4::dot);
	RegistFunc(L, "cross", lua_vector4::cross);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "matrix22");
	lua_newtable(L);
	RegistFunc(L, "new", lua_matrix22::_new);
	RegistFunc(L, "assign", lua_matrix22::assign);
	RegistFunc(L, "mul", lua_matrix22::mul);
	RegistFunc(L, "mulvec", lua_matrix22::mulvec);
	RegistFunc(L, "identity", lua_matrix22::identity);
	RegistFunc(L, "invert", lua_matrix22::invert);
	RegistFunc(L, "scale", lua_matrix22::scale);
	RegistFunc(L, "rotate", lua_matrix22::rotate);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "matrix33");
	lua_newtable(L);
	RegistFunc(L, "new", lua_matrix33::_new);
	RegistFunc(L, "assign", lua_matrix33::assign);
	RegistFunc(L, "mul", lua_matrix33::mul);
	RegistFunc(L, "mulvec", lua_matrix33::mulvec);
	RegistFunc(L, "identity", lua_matrix33::identity);
	RegistFunc(L, "invert", lua_matrix33::invert);
	RegistFunc(L, "scale", lua_matrix33::scale);
	RegistFunc(L, "rotate", lua_matrix33::rotate);
	RegistFunc(L, "rotateX", lua_matrix33::rotateX);
	RegistFunc(L, "rotateY", lua_matrix33::rotateY);
	RegistFunc(L, "rotateZ", lua_matrix33::rotateZ);
	RegistFunc(L, "rotateA", lua_matrix33::rotateA);
	RegistFunc(L, "aim_vector", lua_matrix33::aim_vector);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "matrix44");
	lua_newtable(L);
	RegistFunc(L, "new", lua_matrix44::_new);
	RegistFunc(L, "assign", lua_matrix44::assign);
	RegistFunc(L, "mul", lua_matrix44::mul);
	RegistFunc(L, "mulvec", lua_matrix44::mulvec);
	RegistFunc(L, "identity", lua_matrix44::identity);
	RegistFunc(L, "invert", lua_matrix44::invert);
	RegistFunc(L, "translate", lua_matrix44::translate);
	RegistFunc(L, "scale", lua_matrix44::scale);
	RegistFunc(L, "rotate", lua_matrix44::rotate);
	RegistFunc(L, "rotateX", lua_matrix44::rotateX);
	RegistFunc(L, "rotateY", lua_matrix44::rotateY);
	RegistFunc(L, "rotateZ", lua_matrix44::rotateZ);
	RegistFunc(L, "rotateA", lua_matrix44::rotateA);
	RegistFunc(L, "aim_vector", lua_matrix44::aim_vector);
	lua_settable(L, LUA_GLOBALSINDEX);


}

#undef RegistFunc





} // lua
} // ist

