#ifndef _RGBA_INCLUDE_
#define _RGBA_INCLUDE_


namespace ist {

typedef unsigned char u_char;


template <typename T>
struct RGBA {
	union {
		struct { T r,g,b,a; };
		T v[4];
	};

	RGBA<T> operator + (RGBA<T> &t) const { return RGBA<T>(r+t.r, g+t.g, b+t.b, a+t.a ); }
	RGBA<T> operator - (RGBA<T> &t) const { return RGBA<T>(r-t.r, g-t.g, b-t.b, a-t.a ); }
	RGBA<T> operator * (RGBA<T> &t) const { return RGBA<T>(r*t.r, g*t.g, b*t.b, a*t.a ); }
	RGBA<T> operator / (RGBA<T> &t) const { return RGBA<T>(r/t.r, g/t.g, b/t.b, a/t.a ); }
	RGBA<T>& operator +=(RGBA<T> &t) { *this=*this+t; return *this; }
	RGBA<T>& operator -=(RGBA<T> &t) { *this=*this-t; return *this; }
	RGBA<T>& operator *=(RGBA<T> &t) { *this=*this*t; return *this; }
	RGBA<T>& operator /=(RGBA<T> &t) { *this=*this/t; return *this; }
	bool operator ==(const RGBA<T> &t) const { return (r==t.r && g==t.g && b==t.b && a==t.a); }
	bool operator !=(const RGBA<T> &t) const { return !(*this==t); }
	T& operator [](int i) { return v[i]; }
	const T& operator [](int i) const { return v[i]; }

	RGBA<T>(T _r=0, T _g=0, T _b=0, T _a=0) { r=_r; g=_g; b=_b; a=_a; }

	T shin() const { return (r*0.299 + g*0.587 + b*0.114); }
	RGBA<float> BtoF() const { return RGBA<float>((float)r/255, (float)g/255, (float)b/255, (float)a/255 ); }
	RGBA<u_char>  FtoB() const { return RGBA<u_char>((u_char)(r*255), (u_char)(g*255), (u_char)(b*255), (u_char)(a*255) ); }
};


typedef RGBA<u_char> bRGBA;
typedef RGBA<float> fRGBA;


}	// ist

#endif
