#include <algorithm>
#include "ist_lua.h"
#include "timeline.h"
using namespace std;
using namespace ist;



template<class T>
struct less_x {
	bool operator()(const T& lhs, const T&rhs) {
		return lhs.x < rhs.x;
	}
};


Timeline1::Timeline1(float r) :
	rate(r)
{}

void Timeline1::sampling()
{
	s_points.clear();

	for(iterator p=begin(); p!=end(); ++p) {
		if(p==end()-1)
			break;

		for(float t=p->pos.x; t<=(p+1)->pos.x; t+=rate) {
			s_points.push_back( bezier(t) );
		}
	}

	if(size()==1)
		s_points.push_back( begin()->pos );

	sort(s_points.begin(), s_points.end(), less_x<vector2>());
}

float Timeline1::getValue(float t)
{
	if(s_points.empty())
		return 0;
	else if(t <= s_points.front().x)
		return s_points.front().y;
	else if(t >= s_points.back().x)
		return s_points.back().y;

	typedef vector<vector2>::iterator v2iter;
	pair<v2iter, v2iter> range=equal_range(s_points.begin(), s_points.end(), vector2(t, 0), less_x<vector2>());
	--range.first;
	float dir=(range.second->y - range.first->y)/(range.second->x - range.first->x);
	return range.first->y + dir*(t-range.first->x);
}

void Timeline1::erase(iterator itr)
{
	ist::Spline2D::erase(itr);
	sampling();
}

void Timeline1::insert(const vector2& p)
{
	ist::Spline2D::insert(p);
	sampling();
}

void Timeline1::insert(const vector2& i, const vector2& p, const vector2& o)
{
	ist::Spline2D::insert(i, p, o);
	sampling();
}



Timeline3::Timeline3(float r) :
	x(r), y(r), z(r)
{}

void Timeline3::sampling()
{
	x.sampling();
	y.sampling();
	z.sampling();
}

vector3 Timeline3::getValue(float t)
{
	return vector3(x.getValue(t), y.getValue(t), z.getValue(t));
}







namespace timeline {

unsigned int old_math3d_stack=0;
Timeline3 *current_tl3=0;
Timeline1 *current_tl1=0;
string last_label;



inline void spline_push_point(Timeline1& spline, lua_State *L)
{
	if(lua_gettop(L)==1 && lua_islightuserdata(L, 1)) {
		spline.insert( lua::tovector2(L, 1) );
	}
	else if(lua_gettop(L)==3 && lua_islightuserdata(L, 1) && lua_islightuserdata(L, 2) && lua_islightuserdata(L, 3)) {
		spline.insert( lua::tovector2(L, 1), lua::tovector2(L, 2), lua::tovector2(L, 3) );
	}
	else {
		lua::throw_error(L);
	}
}



int timeline3_begin(lua_State *L)
{
	if(!!current_tl3)
		throw Error("timeline3_begin()");

	if(lua_isstring(L, 1)) {
		old_math3d_stack=lua::math3d::stack_size();
		last_label=lua_tostring(L, 1);
		current_tl3=new Timeline3();
	}
	return 0;
}

int timeline3_insert_x(lua_State *L)
{
	spline_push_point(current_tl3->x, L);
	return 0;
}

int timeline3_insert_y(lua_State *L)
{
	spline_push_point(current_tl3->y, L);
	return 0;
}

int timeline3_insert_z(lua_State *L)
{
	spline_push_point(current_tl3->z, L);
	return 0;
}

int timeline3_end(lua_State *L)
{
	current_tl3->sampling();
	add_timeline3(last_label, current_tl3);

	lua::math3d::stack_resize(old_math3d_stack);
	last_label="";
	current_tl3=0;

	return 0;
}



int timeline1_begin(lua_State *L)
{
	if(!!current_tl1)
		throw Error("timeline1_begin()");

	if(lua_isstring(L, 1)) {
		old_math3d_stack=lua::math3d::stack_size();
		last_label=lua_tostring(L, 1);
		current_tl1=new Timeline1();
	}
	return 0;
}

int timeline1_push(lua_State *L)
{
	spline_push_point(*current_tl1, L);
	return 0;
}

int timeline1_end(lua_State *L)
{
	current_tl1->sampling();
	add_timeline1(last_label, current_tl1);

	lua::math3d::stack_resize(old_math3d_stack);
	last_label="";
	current_tl1=0;

	return 0;
}




void regist_timeline_functions(lua_State *L)
{
	lua::math3dlibopen(L);

	lua_register(L, "timeline3_begin", timeline3_begin);
	lua_register(L, "timeline3_insert_x", timeline3_insert_x);
	lua_register(L, "timeline3_insert_y", timeline3_insert_y);
	lua_register(L, "timeline3_insert_z", timeline3_insert_z);
	lua_register(L, "timeline3_end", timeline3_end);

	lua_register(L, "timeline1_begin", timeline1_begin);
	lua_register(L, "timeline1_push", timeline1_push);
	lua_register(L, "timeline1_end", timeline1_end);
}


} // namespace timeline

